import { HASHTAG_FILTER, URL_FILTER, YOUTUBE_API_KEY } from '$env/static/private';
import type { Post, Tag, TimelineEvent } from '$lib/mastodon/response';
import { savePost } from '$lib/server/db';
import { WebSocket } from "ws";

const YOUTUBE_REGEX = new RegExp('https?:\/\/(www\.)?youtu((be.com\/.*v=)|(\.be\/))(?<videoId>[a-zA-Z_0-9-]+)');

export class TimelineReader {
  private static _instance: TimelineReader;

  private static async isMusicVideo(videoId: string) {
    const searchParams = new URLSearchParams([
      ['part', 'snippet'],
      ['id', videoId],
      ['key', YOUTUBE_API_KEY]]);
    const youtubeVideoUrl = new URL(`https://www.googleapis.com/youtube/v3/videos?${searchParams}`);
    const resp = await fetch(youtubeVideoUrl);
    const respObj = await resp.json();
    const item = respObj.items[0];
    if (item.tags?.includes('music')) {
      return true;
    }
    const categorySearchParams = new URLSearchParams([
      ['part', 'snippet'],
      ['id', item.categoryId],
      ['key', YOUTUBE_API_KEY]]);
    const youtubeCategoryUrl = new URL(`https://www.googleapis.com/youtube/v3/videoCategories?${categorySearchParams}`);
    const categoryTitle: string = await fetch(youtubeCategoryUrl).then(r => r.json()).then(r => r.items[0]?.title);
    return categoryTitle === 'Music';
  }

  private constructor() {
    //const socket = new WebSocket("wss://metalhead.club/api/v1/streaming/public/local")
    const socket = new WebSocket("wss://metalhead.club/api/v1/streaming")
    socket.onopen = (_event) => {
      socket.send('{ "type": "subscribe", "stream": "public:local"}');
    };
    socket.onmessage = (async (event) => {
      try {
        const data: TimelineEvent = JSON.parse(event.data.toString());
        if (data.event !== 'update') {
          return;
        }
        const post: Post = JSON.parse(data.payload);
        const hashttags: string[] = HASHTAG_FILTER.split(',');
        const found_tags: Tag[] = post.tags.filter((t: Tag) => hashttags.includes(t.name));

        const urls: string[] = URL_FILTER.split(',');
        const found_urls = urls.filter(t => post.content.includes(t));

        if (found_urls.length === 0 && found_tags.length === 0) {
          return;
        }

        const youtubeMatches = found_urls.map(u => u.match(YOUTUBE_REGEX)).filter(i => i !== null);
        let hasMusicLink = false;
        // Only check url if no tags are present and if no other matched URLs (e.g. to bandcamp) are found
        if (found_tags.length === 0 && found_urls.length > 0 && youtubeMatches.length === found_urls.length) {
          for (let match of youtubeMatches) {
            if (match === null) {
              continue;
            }
            try {
              const isMusic = await TimelineReader.isMusicVideo(match.groups?.videoId ?? '');
              if (isMusic) {
                hasMusicLink = true;
                break;
              }
            } catch (e) {
              console.error('Could not check if', youtubeMatches, 'is a music video', e);
            }
          }
          if (!hasMusicLink) {
            console.info('Found youtube urls, but none is music', youtubeMatches);
            return;
          }
        }
        savePost(post);

      } catch (e) {
        console.error("error message", event, event.data, e)
      }

    });
    socket.onclose = (event) => {
      console.log("Closed", event, event.code, event.reason)
    };
    socket.onerror = (event) => {
      console.log("error", event, event.message, event.error)
    };

  }

  public static init() {
    if (this._instance === undefined) {
      this._instance = new TimelineReader();
    }
  }

  public static get instance(): TimelineReader {
    TimelineReader.init();
    return this._instance;
  }
}